﻿using System;
using System.ComponentModel;
using System.Windows;
using ScaraBattle.Model;
using ScaraBattle.Persistence;
using ScaraBattle.View;
using ScaraBattle.ViewModel;
using Microsoft.Win32;
using System.Windows.Input;
using System.Windows.Controls;

namespace ScaraBattle
{
    /// <summary>
    /// Interaction logic for App.xaml
    /// </summary>
    public partial class App : Application
    {
        #region Fields

        private GameModel _model;
        private GameViewModel _viewModel;
        private MainWindow _view;

        private SizeDialogBox _sizeDialogBox;

        private bool _over = true;

        #endregion

        #region Constructors

        /// <summary>
        /// Instantiate application
        /// </summary>
        public App()
        {
            Startup += new StartupEventHandler(App_Startup);
        }

        #endregion

        #region Application event handlers

        private void App_Startup(object sender, StartupEventArgs e)
        {
            _model = new GameModel(new GameFileDataAccess());
            _model.GameOver += new EventHandler<GameEventArgs>(Model_GameOver);

            _viewModel = new GameViewModel(_model);
            _viewModel.NewGame += new EventHandler(ViewModel_NewGame);
            _viewModel.CreateMap += new EventHandler(ViewModel_CreateMap);
            _viewModel.ExitGame += new EventHandler(ViewModel_ExitGame);
            _viewModel.LoadGame += new EventHandler(ViewModel_LoadGame);
            _viewModel.SaveGame += new EventHandler(ViewModel_SaveGame);
            _viewModel.Help += new EventHandler(ViewModel_RequestHelp);
            _viewModel.EditMap += new EventHandler(ViewModel_EditMap);

            EventManager.RegisterClassHandler(typeof(Window), Keyboard.KeyUpEvent, new KeyEventHandler(KeyUp), true);

            _view = new MainWindow();
            _view.DataContext = _viewModel;
            _view.Closing += new CancelEventHandler(View_Closing);
            _view.Show();
        }

        #endregion

        #region View event handlers

        /// <summary>
        /// Event handler for closing the game
        /// </summary>
        private void View_Closing(object sender, CancelEventArgs e)
        {
            if (MessageBox.Show("Biztosan ki szeretne lépni?", "ScaraBattle", MessageBoxButton.YesNo, MessageBoxImage.Question) == MessageBoxResult.No)
            {
                e.Cancel = true;
            }
        }

        /// <summary>
        /// Event handler for pressing keys
        /// </summary>
        private void KeyUp(object sender, KeyEventArgs e)
        {
            if (!_over)
            {
                switch (e.Key)
                {
                    case Key.A:
                        _viewModel.OnCameraLeft();
                        break;
                    case Key.S:
                        _viewModel.OnCameraDown();
                        break;
                    case Key.D:
                        _viewModel.OnCameraRight();
                        break;
                    case Key.W:
                        _viewModel.OnCameraUp();
                        break;
                }
            }
        }

        #endregion

        #region ViewModel event handlers

        /// <summary>
        /// Event handler for creating a new game
        /// </summary>
        private void ViewModel_NewGame(object sender, EventArgs e)
        {
            _sizeDialogBox = new SizeDialogBox();

            _sizeDialogBox.ShowDialog();

            if (_sizeDialogBox.DialogResult.HasValue && _sizeDialogBox.DialogResult.Value)
            {
                _over = false;

                int width = _sizeDialogBox.MapWidth;
                int height = _sizeDialogBox.MapHeight;
                int barracks = _sizeDialogBox.Barracks;
                int terrain = _sizeDialogBox.Terrain;

                _model.GenerateMap(width, height, barracks, terrain);
            }
        }

        /// <summary>
        /// Event handler for creating a new map
        /// </summary>
        private void ViewModel_CreateMap(object sender, EventArgs e)
        {
            string mapError = _model.CheckMap();

            if (mapError == String.Empty)
            {
                // nincs hiba, elindíthatjuk a játékot
                _model.InitializeFromMapEditor();
            }
            else
            {
                MessageBox.Show(mapError, "Hiba!", MessageBoxButton.OK, MessageBoxImage.Error);
            }
        }

        /// <summary>
        /// Event handler for loading new game
        /// </summary>
        private async void ViewModel_LoadGame(object sender, EventArgs e)
        {
            try
            {
                OpenFileDialog openFileDialog = new OpenFileDialog();
                openFileDialog.Title = "ScaraBattle tábla betöltése";
                openFileDialog.Filter = "ScaraBattle File|*.sbf";
                if (openFileDialog.ShowDialog() == true)
                {
                    await _model.LoadGameAsync(openFileDialog.FileName);
                    if (_model.CheckMap() != string.Empty)
                    {
                        throw new GameDataException();
                    }


                    _over = false;
                }
            }
            catch (GameDataException)
            {
                MessageBox.Show("Játék betöltése sikertelen!\nHibás az elérési út, vagy a fájlformátum.", "Hiba!", MessageBoxButton.OK, MessageBoxImage.Error);
            }
        }

        /// <summary>
        /// Event handler for saving game
        /// </summary>
        private async void ViewModel_SaveGame(object sender, EventArgs e)
        {
            if (_over)
            {
                MessageBox.Show("Nincs aktív játék, amit el lehetne menteni!", "Hiba!", MessageBoxButton.OK, MessageBoxImage.Error);
                return;
            } else if (_model.Table.State == GameState.Simulation )
            {
                MessageBox.Show("Szimuláció közben nem lehet játékot menteni!", "Hiba!", MessageBoxButton.OK, MessageBoxImage.Error);
                return;
            }

            try
            {
                string mapError = _model.CheckMap();

                if (mapError == String.Empty)
                {
                    SaveFileDialog saveFileDialog = new SaveFileDialog();
                    saveFileDialog.Title = "ScaraBattle tábla betöltése";
                    saveFileDialog.Filter = "ScaraBattle File|*.sbf";
                    if (saveFileDialog.ShowDialog() == true)
                    {
                        try
                        {
                            await _model.SaveGameAsync(saveFileDialog.FileName);
                            MessageBox.Show("Sikeres mentés.", "Mentés", MessageBoxButton.OK, MessageBoxImage.Information);
                        }
                        catch (GameDataException)
                        {
                            MessageBox.Show("Játék mentése sikertelen!\nHibás az elérési út, vagy a könyvtár nem írható.", "Hiba!", MessageBoxButton.OK, MessageBoxImage.Error);
                        }
                    }
                }
                else
                {
                    MessageBox.Show(mapError, "Hiba!", MessageBoxButton.OK, MessageBoxImage.Error);
                }



            }
            catch
            {
                MessageBox.Show("A fájl mentése sikertelen!", "ScaraBattle", MessageBoxButton.OK, MessageBoxImage.Error);
            }
        }

        /// <summary>
        /// Event handler for exiting game
        /// </summary>
        private void ViewModel_ExitGame(object sender, EventArgs e)
        {
            _view.Close(); // ablak bezárása
        }

        /// <summary>
        /// Method for telling the player the rules of the game
        /// </summary>ViewModel_RequestHelp
        private void ViewModel_RequestHelp(object sender, EventArgs e)
        {
            MessageBox.Show("A játék célja, hogy lerombold az ellenfeled kastályát.\n" +
                "A játék körökből áll, a körök pedig építkezés és támadás fázisból.\n\n" +
                "Építkezés:\nKépezz egységeket (amennyiben van elég aranyad), amik a legrövidebb úton megpróbálnak eljutni " +
                "az ellenfeled kastélyához az akadályokat kikerülve, hogy megsebezzék. " +
                "Építs tornyokat, amik elpusztítják az ellenfeled közeledő egységeit. " +
                "A tornyaidat fejlesztheted vagy le is bonthatod.\n\n" +
                "Támadás:\nAz egységek automatikusan mozognak adott számú mezőt, a " +
                "tornyok tüzet nyitnak a hatókörükben lévő valamely ellenséges egységre, amivel elpusztítják azt. " +
                "Amennyiben egy egység rálép az ellenfél kastélyára, megsemmisül és csökkenti a kastély életerejét, " +
                "ami ha nullára csökkent, akkor veszített a tulajdonosa. Ha egy körben mindkét játékos kastélya összeomlik, " +
                "akkor az eredmény döntetlen. Több egység is tartózkodhat egy mezőn (akár ellentétes játékosoké is), " +
                "az egységek nem bocsátkoznak harcba egymással. \n\n" +
                "Pénzszerzés:\n" +
                "Minden kör elején egy adott összeg + korábban kiképzett egységekért + előző körben elpusztított ellenséges egységekért.\n\n" +
                "Megjelenítés:\n" +
                "Bármely cellára kattintva (ha előtte nincs kiválasztva pl. építeni kívánt torony) " +
                "megjelenik alul a cella tartalma, pl több egység esetén, hogy melyik játékosnak hány egysége tartózkodik ott, " +
                "vagy tornyok esetén a részletek (pl hatótávolság).\n" +
                "A \"kamerát\" a WASD billentyűkkel vagy a pálya 4 oldalán található gombokkal mozgathatod és nézheted meg a pálya éppen nem látható részeit.\n\n" +
                "Térképszerkesztő:\n" +
                "Lehetőséged van a random generált pályák mellett sajátot is létrehozni, amit elmenthetsz akár későbbi játékhoz is. " +
                "A választható mennyiségű akadályokon kívül megfelelő mennyiségű barakkot és kastélyt kell elhelyezned mindkét játékosnak, " +
                "enélkül nem lehet elindítani a játékot. Ha kastélyt vagy barakkot szeretnél építeni, először kék kerül a pályára, " +
                "majd ha mégegyszer kattintasz rá (vagy később ugyanolyan objektummal cserélnéd le), " +
                "akkor átvált pirosra (legközelebb meg újra kékre, és így tovább). " +
                "Egy már korábban elmentett saját térképet újra megnyitva akár tovább is szerkeszthetsz.", "ScaraBattle",
                MessageBoxButton.OK, MessageBoxImage.Information);
        }

        /// <summary>
        /// Method opening the map editor
        /// </summary>
        private void ViewModel_EditMap(object sender, EventArgs e)
        {
            _sizeDialogBox = new SizeDialogBox(true);

            _sizeDialogBox.ShowDialog();

            if (_sizeDialogBox.DialogResult.HasValue && _sizeDialogBox.DialogResult.Value)
            {
                _over = false;

                int width = _sizeDialogBox.MapWidth;
                int height = _sizeDialogBox.MapHeight;

                _model.GenerateMap(width, height);
            }
        }

        #endregion

        #region Model event handlers

        /// <summary>
        /// Event handler for game over state
        /// </summary>
        private void Model_GameOver(object sender, GameEventArgs e)
        {
            _viewModel.RefreshTable();

            if (e.WhoWon == 1) // győztestől függő üzenet megjelenítése
            {
                MessageBox.Show("A kék játékos nyert!");
            }
            else
            {
                if (e.WhoWon == 2)
                {
                    MessageBox.Show("A piros játékos nyert!");
                }
                else
                {
                    MessageBox.Show("Döntetlen!");
                }
            }

            _over = true;
        }

        #endregion
    }
}
