using System;
using System.Collections.Generic;
using System.Threading.Tasks;
using ScaraBattle.Model;
using ScaraBattle.Persistence;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Moq;

namespace ScaraBattleTest
{
    [TestClass]
    public class GameTest
    {
        private GameModel _model;
        private GameTable _mockedTable;
        private Mock<IGameDataAccess> _mock;

        [TestInitialize]
        public void Initialize()
        {
            _mockedTable = new GameTable(10, 10, GameState.Menu, 100);

            for (int i = 1; i <= 3; ++i)
            {
                for (int j = 1; j <= 3; ++j)
                {
                    _mockedTable[i, j] = Dune.Instance;
                }
            }

            for (int i = 5; i <= 6; ++i)
            {
                for (int j = 0; j <= 4; ++j)
                {
                    _mockedTable[i, j] = Oasis.Instance;
                }
            }

            for (int i = 8; i <= 9; ++i)
            {
                for (int j = 4; j <= 7; ++j)
                {
                    _mockedTable[i, j] = Ruins.Instance;
                }
            }

            _mockedTable.BlueCastle = new ScaraBattle.Persistence.Castle(10) { IsBlue = true };
            _mockedTable[4, 0] = _mockedTable.BlueCastle;
            _mockedTable.BlueCastle.Position = (4, 0);

            _mockedTable.BlueBarracks = new (int, int)[2];
            _mockedTable[0, 0] = new Barrack() { IsBlue = true };
            _mockedTable.BlueBarracks[0] = (0, 0);
            _mockedTable[8, 0] = new Barrack() { IsBlue = true };
            _mockedTable.BlueBarracks[1] = (8, 0);


            _mockedTable.RedCastle = new ScaraBattle.Persistence.Castle(10) { IsBlue = false };
            _mockedTable[0, 9] = _mockedTable.RedCastle;
            _mockedTable.RedCastle.Position = (0, 9);

            _mockedTable.RedBarracks = new (int, int)[2];
            _mockedTable[0, 7] = new Barrack() { IsBlue = false };
            _mockedTable.RedBarracks[0] = (0, 7);
            _mockedTable[9, 9] = new Barrack() { IsBlue = false };
            _mockedTable.RedBarracks[1] = (9, 9);

            // Mocked table:
            // B . . . . . . b . c
            // . D D D . . . . . .
            // . D D D . . . . . .
            // . D D D . . . . . .
            // C . . . . . . . . .
            // O O O O O . . . . .
            // O O O O O . . . . .
            // . . . . . . . . . .
            // B . . . R R R R . .
            // . . . . R R R R . b

            // Notation:
            // . - Empty space
            // D - Dune
            // O - Oasis
            // R - Ruins
            // C - Blue castle
            // B - Blue barrack
            // c - Red castle
            // b - Red barrack

            _mock = new Mock<IGameDataAccess>();
            _mock.Setup(mock => mock.LoadAsync(It.IsAny<String>()))
                .Returns(() => Task.FromResult(_mockedTable));

            _model = new GameModel(_mock.Object);
        }

        [TestMethod, TestCategory("Trivial")]
        public async Task GameTablePropertyInitialization()
        {
            await _model.LoadGameAsync(String.Empty);

            Assert.AreEqual(10, _model.Table.Height);
            Assert.AreEqual(10, _model.Table.Width);

            Assert.AreEqual(100, _model.Table.BlueGold);
            Assert.AreEqual(10, _model.Table.BlueHealth);

            Assert.AreEqual(100, _model.Table.RedGold);
            Assert.AreEqual(10, _model.Table.RedHealth);

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("Trivial")]
        public async Task GameTableObjectInitialization()
        {
            await _model.LoadGameAsync(String.Empty);

            for (int i = 1; i <= 3; ++i)
            {
                for (int j = 1; j <= 3; ++j)
                {
                    Assert.AreEqual(Dune.Instance, _model.Table[i, j]);
                }
            }

            for (int i = 5; i <= 6; ++i)
            {
                for (int j = 0; j <= 4; ++j)
                {
                    Assert.AreEqual(Oasis.Instance, _model.Table[i, j]);
                }
            }

            for (int i = 8; i <= 9; ++i)
            {
                for (int j = 4; j <= 7; ++j)
                {
                    Assert.AreEqual(Ruins.Instance, _model.Table[i, j]);
                }
            }

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("Trivial")]
        public async Task GameTableCastleInitialization()
        {
            await _model.LoadGameAsync(String.Empty);

            Assert.IsTrue(_mockedTable[4, 0] is ScaraBattle.Persistence.Castle && ((ScaraBattle.Persistence.Castle)_mockedTable[4, 0]).IsBlue);
            Assert.AreEqual((4, 0), _mockedTable.BlueCastle.Position);

            Assert.IsTrue(_mockedTable[0, 9] is ScaraBattle.Persistence.Castle && !((ScaraBattle.Persistence.Castle)_mockedTable[0, 9]).IsBlue);
            Assert.AreEqual((0, 9), _mockedTable.RedCastle.Position);

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("Trivial")]
        public async Task GameTableBarrackInitialization()
        {
            await _model.LoadGameAsync(String.Empty);

            Assert.IsTrue(_mockedTable[0, 0] is Barrack && ((Barrack)_mockedTable[0, 0]).IsBlue);
            Assert.AreEqual((0, 0), _mockedTable.BlueBarracks[0]);
            Assert.IsTrue(_mockedTable[8, 0] is Barrack && ((Barrack)_mockedTable[8, 0]).IsBlue);
            Assert.AreEqual((8, 0), _mockedTable.BlueBarracks[1]);

            Assert.IsTrue(_mockedTable[0, 7] is Barrack && !((Barrack)_mockedTable[0, 7]).IsBlue);
            Assert.AreEqual((0, 7), _mockedTable.RedBarracks[0]);
            Assert.IsTrue(_mockedTable[9, 9] is Barrack && !((Barrack)_mockedTable[9, 9]).IsBlue);
            Assert.AreEqual((9, 9), _mockedTable.RedBarracks[1]);

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("GenerateMap")]
        public async Task GenerateMapRandomly()
        {
            Random r = new Random();

            for (int i = 1; i <= 100; ++i)
            {
                _model.GenerateMap(r.Next(10, 21), r.Next(10, 21), r.Next(2, 4), r.Next(0, 3));
                Assert.AreEqual(String.Empty, _model.CheckMap());
            }
        }

        [TestMethod, TestCategory("StepGameState"), TestCategory("Gameplay")]
        public async Task StepGameStateFromBlue()
        {
            await _model.LoadGameAsync(String.Empty);

            _model.Table.State = GameState.Blue;
            _model.StepGameState();
            Assert.AreEqual(GameState.Red, _model.Table.State);

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("StepGameState"), TestCategory("Gameplay")]
        public async Task StepGameStateFromRed()
        {
            await _model.LoadGameAsync(String.Empty);

            _model.Table.State = GameState.Red;
            _model.StepGameState();
            Assert.AreEqual(GameState.Simulation, _model.Table.State);

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("StepGameState")]
        public async Task StepGameStateNoChange()
        {
            await _model.LoadGameAsync(String.Empty);

            Assert.AreEqual(GameState.Menu, _model.Table.State);
            _model.StepGameState();
            Assert.AreEqual(GameState.Menu, _model.Table.State);

            _model.Table.State = GameState.MapEditor;
            _model.StepGameState();
            Assert.AreEqual(GameState.MapEditor, _model.Table.State);

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("Connectivity")]
        public async Task ConnectivityCheckCommutativity()
        {
            await _model.LoadGameAsync(String.Empty);

            Assert.IsTrue(_model.ConnectivityCheck(9, 9, 0, 0));
            Assert.IsTrue(_model.ConnectivityCheck(0, 0, 9, 9));
            _model.Table[0, 3] = Ruins.Instance;
            Assert.IsFalse(_model.ConnectivityCheck(9, 9, 0, 0));
            Assert.IsFalse(_model.ConnectivityCheck(0, 0, 9, 9));

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("Connectivity")]
        public async Task ConnectivityCheckDifferentObjects()
        {
            await _model.LoadGameAsync(String.Empty);

            Assert.IsTrue(_model.ConnectivityCheck(8, 0, 0, 9));
            _model.Table[7, 4] = Dune.Instance;
            Assert.IsFalse(_model.ConnectivityCheck(8, 0, 0, 9));
            _model.Table[7, 4] = Oasis.Instance;
            Assert.IsFalse(_model.ConnectivityCheck(8, 0, 0, 9));
            _model.Table[7, 4] = Ruins.Instance;
            Assert.IsFalse(_model.ConnectivityCheck(8, 0, 0, 9));
            _model.Table[7, 4] = new ScaraBattle.Persistence.Castle();
            Assert.IsFalse(_model.ConnectivityCheck(8, 0, 0, 9));
            _model.Table[7, 4] = new Barrack();
            Assert.IsFalse(_model.ConnectivityCheck(8, 0, 0, 9));
            _model.Table[7, 4] = new SmallTower();
            Assert.IsFalse(_model.ConnectivityCheck(8, 0, 0, 9));
            _model.Table[7, 4] = new MediumTower();
            Assert.IsFalse(_model.ConnectivityCheck(8, 0, 0, 9));
            _model.Table[7, 4] = new LargeTower();
            Assert.IsFalse(_model.ConnectivityCheck(8, 0, 0, 9));
            _model.Table[7, 4] = null;
            Assert.IsTrue(_model.ConnectivityCheck(8, 0, 0, 9));

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("InitializeFromMapEditor"), TestCategory("MapEditor")]
        public async Task InitializeFromMapEditorTest()
        {
            await _model.LoadGameAsync(String.Empty);

            _model.GenerateMap(20, 20);

            Assert.AreEqual(GameState.MapEditor, _model.Table.State);

            for (int i = 1; i <= 3; ++i)
            {
                for (int j = 1; j <= 3; ++j)
                {
                    _model.Table[i, j] = Dune.Instance;
                }
            }

            for (int i = 5; i <= 6; ++i)
            {
                for (int j = 0; j <= 4; ++j)
                {
                    _model.Table[i, j] = Oasis.Instance;
                }
            }

            for (int i = 8; i <= 9; ++i)
            {
                for (int j = 4; j <= 7; ++j)
                {
                    _model.Table[i, j] = Ruins.Instance;
                }
            }

            _model.Table[4, 0] = new ScaraBattle.Persistence.Castle() { IsBlue = true };

            _model.Table[0, 0] = new Barrack() { IsBlue = true };
            _model.Table[8, 0] = new Barrack() { IsBlue = true };


            _model.Table[0, 9] = new ScaraBattle.Persistence.Castle() { IsBlue = false };

            _model.Table[0, 7] = new Barrack() { IsBlue = false };
            _model.Table[9, 9] = new Barrack() { IsBlue = false };

            Assert.AreEqual(String.Empty, _model.CheckMap());

            _model.InitializeFromMapEditor();


            Assert.IsTrue(_model.Table[4, 0] is ScaraBattle.Persistence.Castle && ((ScaraBattle.Persistence.Castle)_model.Table[4, 0]).IsBlue);
            Assert.AreEqual((4, 0), _model.Table.BlueCastle.Position);

            Assert.IsTrue(_model.Table[0, 0] is Barrack && ((Barrack)_model.Table[0, 0]).IsBlue);
            Assert.AreEqual((0, 0), _model.Table.BlueBarracks[0]);
            Assert.IsTrue(_model.Table[8, 0] is Barrack && ((Barrack)_model.Table[8, 0]).IsBlue);
            Assert.AreEqual((8, 0), _model.Table.BlueBarracks[1]);


            Assert.IsTrue(_model.Table[0, 9] is ScaraBattle.Persistence.Castle && !((ScaraBattle.Persistence.Castle)_model.Table[0, 9]).IsBlue);
            Assert.AreEqual((0, 9), _model.Table.RedCastle.Position);

            Assert.IsTrue(_model.Table[0, 7] is Barrack && !((Barrack)_model.Table[0, 7]).IsBlue);
            Assert.AreEqual((0, 7), _model.Table.RedBarracks[0]);
            Assert.IsTrue(_model.Table[9, 9] is Barrack && !((Barrack)_model.Table[9, 9]).IsBlue);
            Assert.AreEqual((9, 9), _model.Table.RedBarracks[1]);

            // TODO: Megnzni, hogy a tbla cellja s a kastly tulajdonsg ugyanarra
            // az objektumra hivatkozik-e ilyenkor, mert n nem vagyok benne biztos

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("CheckMap")]
        public async Task CheckMapCorrectMap()
        {
            await _model.LoadGameAsync(String.Empty);

            Assert.AreEqual(String.Empty, _model.CheckMap());

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("CheckMap")]
        public async Task CheckMapConnectivityProblem()
        {
            await _model.LoadGameAsync(String.Empty);

            _model.Table[0, 3] = Dune.Instance;
            Assert.IsTrue(_model.CheckMap().Contains("A trkp nem tjrhat"));
            _model.Table[0, 3] = null;

            _model.Table[7, 4] = Ruins.Instance;
            Assert.IsTrue(_model.CheckMap().Contains("A trkp nem tjrhat"));
            _model.Table[7, 4] = null;

            _model.Table[4, 1] = Oasis.Instance;
            Assert.IsTrue(_model.CheckMap().Contains("A trkp nem tjrhat"));
            _model.Table[4, 1] = null;

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("CheckMap")]
        public async Task CheckMapCastleProblem()
        {
            await _model.LoadGameAsync(String.Empty);

            // 0 kk kastly
            _model.Table[4, 0] = null;
            Assert.IsTrue(_model.CheckMap().Contains("A lehelyezett kk kastlyok szma nem egyenl eggyel"));

            // 2 kk kastly
            _model.Table[4, 0] = new ScaraBattle.Persistence.Castle() { IsBlue = true };
            _model.Table[0, 9] = new ScaraBattle.Persistence.Castle() { IsBlue = true };
            Assert.IsTrue(_model.CheckMap().Contains("A lehelyezett kk kastlyok szma nem egyenl eggyel"));

            // 0 piros kastly
            Assert.IsTrue(_model.CheckMap().Contains("A lehelyezett piros kastlyok szma nem egyenl eggyel"));

            // 2 piros kastly
            _model.Table[4, 0] = new ScaraBattle.Persistence.Castle() { IsBlue = false };
            _model.Table[0, 9] = new ScaraBattle.Persistence.Castle() { IsBlue = false };
            Assert.IsTrue(_model.CheckMap().Contains("A lehelyezett piros kastlyok szma nem egyenl eggyel"));
            _model.Table[0, 9] = new ScaraBattle.Persistence.Castle() { IsBlue = true };

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("CheckMap")]
        public async Task CheckMapBarrackProblem()
        {
            await _model.LoadGameAsync(String.Empty);

            // nem ugyanannyi barakkja van a kt jtkosnak
            _model.Table[9, 9] = null;
            Assert.IsTrue(_model.CheckMap().Contains("A kt jtkos barakkjainak szma nem egyezik meg"));

            // ugyanannyi barakkjuk van, de ez kevesebb mint 2
            _model.Table[0, 0] = null;
            Assert.IsTrue(_model.CheckMap().Contains("Mindkt jtkosnak legalbb 2 barakkal kell rendelkeznie"));

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("MapEditor")]
        public async Task MapEditorPlaceObject()
        {
            await _model.LoadGameAsync(String.Empty);

            _model.Table.State = GameState.MapEditor;

            _model.PlaceObject(5, 5, SelectedObject.Dune);
            Assert.AreEqual(Dune.Instance, _model.Table[5, 5]);

            _model.PlaceObject(5, 5, SelectedObject.Oasis);
            Assert.AreEqual(Oasis.Instance, _model.Table[5, 5]);

            _model.PlaceObject(5, 5, SelectedObject.Ruins);
            Assert.AreEqual(Ruins.Instance, _model.Table[5, 5]);

            _model.PlaceObject(5, 5, SelectedObject.Castle);
            Assert.IsTrue(_mockedTable[5, 5] is ScaraBattle.Persistence.Castle && ((ScaraBattle.Persistence.Castle)_mockedTable[5, 5]).IsBlue);

            _model.PlaceObject(5, 5, SelectedObject.Barrack);
            Assert.IsTrue(_mockedTable[5, 5] is Barrack && ((Barrack)_mockedTable[5, 5]).IsBlue);

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("MapEditor")]
        public async Task MapEditorRemoveObject()
        {
            await _model.LoadGameAsync(String.Empty);

            _model.Table.State = GameState.MapEditor;

            _model.PlaceObject(5, 5, SelectedObject.Nothing);
            Assert.IsNull(_model.Table[5, 5]);

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("MapEditor")]
        public async Task MapEditorChangeCastleColor()
        {
            await _model.LoadGameAsync(String.Empty);

            _model.Table.State = GameState.MapEditor;

            _model.PlaceObject(4, 0, SelectedObject.Castle);
            Assert.IsTrue(_mockedTable[4, 0] is ScaraBattle.Persistence.Castle && !((ScaraBattle.Persistence.Castle)_mockedTable[4, 0]).IsBlue);
            _model.PlaceObject(4, 0, SelectedObject.Castle);
            Assert.IsTrue(_mockedTable[4, 0] is ScaraBattle.Persistence.Castle && ((ScaraBattle.Persistence.Castle)_mockedTable[4, 0]).IsBlue);

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("MapEditor")]
        public async Task MapEditorChangeBarrackColor()
        {
            await _model.LoadGameAsync(String.Empty);

            _model.Table.State = GameState.MapEditor;

            _model.PlaceObject(0, 0, SelectedObject.Barrack);
            Assert.IsTrue(_mockedTable[0, 0] is Barrack && !((Barrack)_mockedTable[0, 0]).IsBlue);
            _model.PlaceObject(0, 0, SelectedObject.Barrack);
            Assert.IsTrue(_mockedTable[0, 0] is Barrack && ((Barrack)_mockedTable[0, 0]).IsBlue);

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("BuyTower"), TestCategory("Gameplay")]
        public async Task BuyDifferentTowerTypes()
        {
            await _model.LoadGameAsync(String.Empty);

            _model.Table.State = GameState.Blue;
            _model.Table.BlueGold = 10000;
            int prevGold = _model.Table.BlueGold;
            Assert.AreEqual(0, _model.Table.BlueTowerPositions.Count);

            _model.PlaceObject(2, 9, SelectedObject.SmallTower);
            Assert.IsTrue(_model.Table[2, 9] is SmallTower && ((SmallTower)_model.Table[2, 9]).IsBlue);
            Assert.AreEqual(prevGold - SmallTower.Price, _model.Table.BlueGold);
            Assert.AreEqual(1, _model.Table.BlueTowerPositions.Count);
            prevGold = _model.Table.BlueGold;

            _model.PlaceObject(3, 9, SelectedObject.MediumTower);
            Assert.IsTrue(_model.Table[3, 9] is MediumTower && ((MediumTower)_model.Table[3, 9]).IsBlue);
            Assert.AreEqual(prevGold - MediumTower.Price, _model.Table.BlueGold);
            Assert.AreEqual(2, _model.Table.BlueTowerPositions.Count);
            prevGold = _model.Table.BlueGold;

            _model.PlaceObject(4, 9, SelectedObject.LargeTower);
            Assert.IsTrue(_model.Table[4, 9] is LargeTower && ((LargeTower)_model.Table[4, 9]).IsBlue);
            Assert.AreEqual(prevGold - LargeTower.Price, _model.Table.BlueGold);
            Assert.AreEqual(3, _model.Table.BlueTowerPositions.Count);
            prevGold = _model.Table.BlueGold;

            _model.Table[2, 9] = null;
            _model.Table[3, 9] = null;
            _model.Table[4, 9] = null;
            _model.Table.BlueTowerPositions = new List<(int, int)>();

            _model.Table.State = GameState.Red;
            _model.Table.RedGold = 10000;
            prevGold = _model.Table.RedGold;
            Assert.AreEqual(0, _model.Table.RedTowerPositions.Count);
            
            _model.PlaceObject(2, 9, SelectedObject.SmallTower);
            Assert.IsTrue(_model.Table[2, 9] is SmallTower && !((SmallTower)_model.Table[2, 9]).IsBlue);
            Assert.AreEqual(prevGold - SmallTower.Price, _model.Table.RedGold);
            Assert.AreEqual(1, _model.Table.RedTowerPositions.Count);
            prevGold = _model.Table.RedGold;

            _model.PlaceObject(3, 9, SelectedObject.MediumTower);
            Assert.IsTrue(_model.Table[3, 9] is MediumTower && !((MediumTower)_model.Table[3, 9]).IsBlue);
            Assert.AreEqual(prevGold - MediumTower.Price, _model.Table.RedGold);
            Assert.AreEqual(2, _model.Table.RedTowerPositions.Count);
            prevGold = _model.Table.RedGold;

            _model.PlaceObject(4, 9, SelectedObject.LargeTower);
            Assert.IsTrue(_model.Table[4, 9] is LargeTower && !((LargeTower)_model.Table[4, 9]).IsBlue);
            Assert.AreEqual(prevGold - LargeTower.Price, _model.Table.RedGold);
            Assert.AreEqual(3, _model.Table.RedTowerPositions.Count);
            prevGold = _model.Table.RedGold;

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("BuyTower"), TestCategory("Gameplay")]
        public async Task BuyTowerPlacementBlocked()
        {
            await _model.LoadGameAsync(String.Empty);

            _model.Table.State = GameState.Blue;
            _model.Table.BlueGold = 10000;
            int prevGold = _model.Table.BlueGold;

            _model.PlaceObject(0, 0, SelectedObject.SmallTower);
            Assert.IsFalse(_model.Table[0, 0] is SmallTower);
            Assert.AreEqual(prevGold, _model.Table.BlueGold);
            Assert.AreEqual(0, _model.Table.BlueTowerPositions.Count);




            _model.Table.State = GameState.Red;
            _model.Table.RedGold = 10000;
            prevGold = _model.Table.RedGold;

            _model.PlaceObject(4, 4, SelectedObject.SmallTower);
            Assert.IsNull(_model.Table[4, 4]);
            Assert.AreEqual(prevGold, _model.Table.RedGold);
            Assert.AreEqual(0, _model.Table.RedTowerPositions.Count);

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("BuyTower"), TestCategory("Gameplay")]
        public async Task BuyTowerConnectivityProblem()
        {
            await _model.LoadGameAsync(String.Empty);

            _model.Table.State = GameState.Blue;
            _model.Table.BlueGold = 10000;
            int prevGold = _model.Table.BlueGold;

            _model.PlaceObject(4, 4, SelectedObject.SmallTower);
            Assert.IsNull(_model.Table[4, 4]);
            Assert.AreEqual(prevGold, _model.Table.BlueGold);
            Assert.AreEqual(0, _model.Table.BlueTowerPositions.Count);




            _model.Table.State = GameState.Red;
            _model.Table.RedGold = 10000;
            prevGold = _model.Table.RedGold;

            _model.PlaceObject(4, 4, SelectedObject.SmallTower);
            Assert.IsNull(_model.Table[4, 4]);
            Assert.AreEqual(prevGold, _model.Table.RedGold);
            Assert.AreEqual(0, _model.Table.RedTowerPositions.Count);

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("BuyTower"), TestCategory("Gameplay")]
        public async Task BuyTowerNotEnoughGold()
        {
            await _model.LoadGameAsync(String.Empty);

            _model.Table.State = GameState.Blue;

            _model.Table.BlueGold = SmallTower.Price - 1;
            _model.PlaceObject(5, 9, SelectedObject.SmallTower);
            Assert.IsNull(_model.Table[5, 9]);
            Assert.AreEqual(SmallTower.Price - 1, _model.Table.BlueGold);
            Assert.AreEqual(0, _model.Table.BlueTowerPositions.Count);




            _model.Table.State = GameState.Red;

            _model.Table.RedGold = SmallTower.Price - 1;
            _model.PlaceObject(5, 9, SelectedObject.SmallTower);
            Assert.IsNull(_model.Table[5, 9]);
            Assert.AreEqual(SmallTower.Price - 1, _model.Table.RedGold);
            Assert.AreEqual(0, _model.Table.RedTowerPositions.Count);

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("SellTower"), TestCategory("Gameplay")]
        public async Task SellTowerInvalidCell()
        {
            await _model.LoadGameAsync(String.Empty);

            _model.Table.State = GameState.Blue;
            _model.PlaceObject(5, 5, SelectedObject.MediumTower);
            _model.PlaceObject(6, 6, SelectedObject.SmallTower);
            Assert.AreEqual(2, _model.Table.BlueTowerPositions.Count);

            _model.Table.State = GameState.Red;
            _model.PlaceObject(9, 1, SelectedObject.LargeTower);
            Assert.AreEqual(1, _model.Table.RedTowerPositions.Count);

            // Nem lehet eladni olyan pozcirl tornyot, amin nincs semmi
            _model.SellTower(7, 7);
            Assert.AreEqual(1, _model.Table.RedTowerPositions.Count);

            // Nem lehet eladni olyan pozcirl tornyot, amin nem torony van
            _model.SellTower(0, 0);
            Assert.AreEqual(1, _model.Table.RedTowerPositions.Count);

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("SellTower"), TestCategory("Gameplay")]
        public async Task SellTowerSuccess()
        {
            await _model.LoadGameAsync(String.Empty);

            _model.Table.State = GameState.Blue;
            _model.PlaceObject(5, 5, SelectedObject.MediumTower);
            _model.PlaceObject(6, 6, SelectedObject.SmallTower);
            Assert.AreEqual(2, _model.Table.BlueTowerPositions.Count);

            _model.Table.State = GameState.Red;
            _model.PlaceObject(9, 1, SelectedObject.LargeTower);
            Assert.AreEqual(1, _model.Table.RedTowerPositions.Count);



            _model.Table.State = GameState.Blue;
            int blueGold = _model.Table.BlueGold;
            int sellCost = ((MediumTower)_model.Table[5, 5]).SellCost;
            _model.SellTower(5, 5);
            Assert.AreEqual(1, _model.Table.BlueTowerPositions.Count);
            Assert.AreEqual(blueGold + sellCost, _model.Table.BlueGold);
            Assert.IsNull(_model.Table[5, 5]);



            int redGold = _model.Table.RedGold;
            _model.Table.State = GameState.Red;
            redGold = _model.Table.RedGold;
            sellCost = ((LargeTower)_model.Table[9, 1]).SellCost;
            _model.SellTower(9, 1);
            Assert.AreEqual(0, _model.Table.RedTowerPositions.Count);
            Assert.AreEqual(redGold + sellCost, _model.Table.RedGold);
            Assert.IsNull(_model.Table[9, 1]);

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("SellTower"), TestCategory("Gameplay")]
        public async Task SellTowerWrongColor()
        {
            await _model.LoadGameAsync(String.Empty);

            _model.Table.State = GameState.Blue;
            _model.PlaceObject(5, 5, SelectedObject.MediumTower);
            _model.PlaceObject(6, 6, SelectedObject.SmallTower);
            Assert.AreEqual(2, _model.Table.BlueTowerPositions.Count);

            _model.Table.State = GameState.Red;
            _model.PlaceObject(9, 1, SelectedObject.LargeTower);
            Assert.AreEqual(1, _model.Table.RedTowerPositions.Count);



            _model.Table.State = GameState.Blue;
            int blueGold = _model.Table.BlueGold;
            int redGold = _model.Table.RedGold;
            _model.SellTower(9, 1);
            Assert.AreEqual(2, _model.Table.BlueTowerPositions.Count);
            Assert.AreEqual(blueGold, _model.Table.BlueGold);
            Assert.AreEqual(1, _model.Table.RedTowerPositions.Count);
            Assert.AreEqual(redGold, _model.Table.RedGold);
            Assert.IsNotNull(_model.Table[9, 1]);


            _model.Table.State = GameState.Red;
            _model.SellTower(6, 6);
            Assert.AreEqual(2, _model.Table.BlueTowerPositions.Count);
            Assert.AreEqual(blueGold, _model.Table.BlueGold);
            Assert.AreEqual(1, _model.Table.RedTowerPositions.Count);
            Assert.AreEqual(redGold, _model.Table.RedGold);
            Assert.IsNotNull(_model.Table[6, 6]);

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("UpgradeTower"), TestCategory("Gameplay")]
        public async Task UpgradeTowerInvalidCell()
        {
            await _model.LoadGameAsync(String.Empty);

            _model.Table.State = GameState.Blue;
            _model.Table.BlueGold = 10000;
            _model.PlaceObject(5, 5, SelectedObject.MediumTower);
            _model.PlaceObject(6, 6, SelectedObject.SmallTower);
            Assert.AreEqual(2, _model.Table.BlueTowerPositions.Count);
            Assert.AreEqual(1, ((SmallTower)_model.Table[6, 6]).Level);
            Assert.AreEqual(4, ((MediumTower)_model.Table[5, 5]).Level);

            _model.Table.State = GameState.Red;
            _model.Table.BlueGold = 10000;
            _model.PlaceObject(9, 1, SelectedObject.LargeTower);
            Assert.AreEqual(1, _model.Table.RedTowerPositions.Count);
            Assert.AreEqual(7, ((LargeTower)_model.Table[9, 1]).Level);

            // Nem lehet fejleszteni olyan pozcin tornyot, amin nincs semmi
            int redGold = _model.Table.RedGold;
            _model.UpgradeTower(7, 7);
            Assert.AreEqual(redGold, _model.Table.RedGold);

            // Nem lehet fejleszteni olyan pozcin tornyot, amin nem torony van
            _model.UpgradeTower(0, 0);
            Assert.AreEqual(redGold, _model.Table.RedGold);

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("UpgradeTower"), TestCategory("Gameplay")]
        public async Task UpgradeTowerSuccess()
        {
            await _model.LoadGameAsync(String.Empty);

            _model.Table.State = GameState.Blue;
            _model.Table.BlueGold = 10000;
            _model.PlaceObject(5, 5, SelectedObject.MediumTower);
            _model.PlaceObject(6, 6, SelectedObject.SmallTower);
            Assert.AreEqual(2, _model.Table.BlueTowerPositions.Count);
            Assert.AreEqual(1, ((SmallTower)_model.Table[6, 6]).Level);
            Assert.AreEqual(4, ((MediumTower)_model.Table[5, 5]).Level);

            _model.Table.State = GameState.Red;
            _model.Table.BlueGold = 10000;
            _model.PlaceObject(9, 1, SelectedObject.LargeTower);
            Assert.AreEqual(1, _model.Table.RedTowerPositions.Count);
            Assert.AreEqual(7, ((LargeTower)_model.Table[9, 1]).Level);

            int redGold = _model.Table.RedGold;
            int blueGold = _model.Table.BlueGold;
            int upgradeCost = ((MediumTower)_model.Table[5, 5]).UpgradeCost;
            _model.Table.State = GameState.Blue;
            _model.UpgradeTower(5, 5);
            Assert.AreEqual(5, ((MediumTower)_model.Table[5, 5]).Level);
            Assert.AreEqual(blueGold - upgradeCost, _model.Table.BlueGold);
            blueGold = _model.Table.BlueGold;




            redGold = _model.Table.RedGold;
            upgradeCost = ((LargeTower)_model.Table[9, 1]).UpgradeCost;
            _model.Table.State = GameState.Red;
            _model.UpgradeTower(9, 1);
            Assert.AreEqual(8, ((LargeTower)_model.Table[9, 1]).Level);
            Assert.AreEqual(redGold - upgradeCost, _model.Table.RedGold);
            redGold = _model.Table.RedGold;

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("UpgradeTower"), TestCategory("Gameplay")]
        public async Task UpgradeTowerWrongColor()
        {
            await _model.LoadGameAsync(String.Empty);

            _model.Table.State = GameState.Blue;
            _model.Table.BlueGold = 10000;
            _model.PlaceObject(5, 5, SelectedObject.MediumTower);
            _model.PlaceObject(6, 6, SelectedObject.SmallTower);
            Assert.AreEqual(2, _model.Table.BlueTowerPositions.Count);
            Assert.AreEqual(1, ((SmallTower)_model.Table[6, 6]).Level);
            Assert.AreEqual(4, ((MediumTower)_model.Table[5, 5]).Level);

            _model.Table.State = GameState.Red;
            _model.Table.BlueGold = 10000;
            _model.PlaceObject(9, 1, SelectedObject.LargeTower);
            Assert.AreEqual(1, _model.Table.RedTowerPositions.Count);
            Assert.AreEqual(7, ((LargeTower)_model.Table[9, 1]).Level);

            int redGold = _model.Table.RedGold;
            int blueGold = _model.Table.BlueGold;



            _model.Table.State = GameState.Blue;
            _model.UpgradeTower(9, 1);
            Assert.AreEqual(7, ((LargeTower)_model.Table[9, 1]).Level);
            Assert.AreEqual(blueGold, _model.Table.BlueGold);




            _model.Table.State = GameState.Red;
            _model.UpgradeTower(6, 6);
            Assert.AreEqual(1, ((SmallTower)_model.Table[6, 6]).Level);
            Assert.AreEqual(redGold, _model.Table.RedGold);

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("UpgradeTower"), TestCategory("Gameplay")]
        public async Task UpgradeTowerNotEnoughMoney()
        {
            await _model.LoadGameAsync(String.Empty);

            _model.Table.State = GameState.Blue;
            _model.Table.BlueGold = 10000;
            _model.PlaceObject(5, 5, SelectedObject.MediumTower);
            _model.PlaceObject(6, 6, SelectedObject.SmallTower);
            Assert.AreEqual(2, _model.Table.BlueTowerPositions.Count);
            Assert.AreEqual(1, ((SmallTower)_model.Table[6, 6]).Level);
            Assert.AreEqual(4, ((MediumTower)_model.Table[5, 5]).Level);

            _model.Table.State = GameState.Red;
            _model.Table.BlueGold = 10000;
            _model.PlaceObject(9, 1, SelectedObject.LargeTower);
            Assert.AreEqual(1, _model.Table.RedTowerPositions.Count);
            Assert.AreEqual(7, ((LargeTower)_model.Table[9, 1]).Level);



            _model.Table.State = GameState.Blue;
            _model.Table.BlueGold = ((MediumTower)_model.Table[5, 5]).UpgradeCost - 1;
            int blueGold = _model.Table.BlueGold;
            _model.UpgradeTower(5, 5);
            Assert.AreEqual(4, ((MediumTower)_model.Table[5, 5]).Level);
            Assert.AreEqual(blueGold, _model.Table.BlueGold);




            _model.Table.State = GameState.Red;
            _model.Table.RedGold = ((LargeTower)_model.Table[9, 1]).UpgradeCost - 1;
            int redGold = _model.Table.RedGold;
            _model.UpgradeTower(9, 1);
            Assert.AreEqual(7, ((LargeTower)_model.Table[9, 1]).Level);
            Assert.AreEqual(redGold, _model.Table.RedGold);

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("UpgradeTower"), TestCategory("Gameplay")]
        public async Task NextLevelTowerTest()
        {
            await _model.LoadGameAsync(String.Empty);

            _model.Table.State = GameState.Blue;
            _model.Table.BlueGold = 10000;

            _model.PlaceObject(6, 6, SelectedObject.SmallTower);
            Assert.IsTrue(_model.Table[6, 6] is SmallTower && ((SmallTower)_model.Table[6, 6]).Level == 1);
            _model.UpgradeTower(6, 6);
            Assert.IsTrue(_model.Table[6, 6] is SmallTower && ((SmallTower)_model.Table[6, 6]).Level == 2);
            _model.UpgradeTower(6, 6);
            _model.UpgradeTower(6, 6);

            Assert.IsTrue(_model.Table[6, 6] is MediumTower && ((MediumTower)_model.Table[6, 6]).Level == 4);
            _model.UpgradeTower(6, 6);
            Assert.IsTrue(_model.Table[6, 6] is MediumTower && ((MediumTower)_model.Table[6, 6]).Level == 5);
            _model.UpgradeTower(6, 6);
            _model.UpgradeTower(6, 6);

            Assert.IsTrue(_model.Table[6, 6] is LargeTower && ((LargeTower)_model.Table[6, 6]).Level == 7);
            for (int i = 1; i <= 10; ++i)
            {
                _model.UpgradeTower(6, 6);
            }
            Assert.IsTrue(_model.Table[6, 6] is LargeTower && ((LargeTower)_model.Table[6, 6]).Level == 17);

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("BuyUnit"), TestCategory("Gameplay")]
        public async Task BuyUnitSuccess()
        {
            await _model.LoadGameAsync(String.Empty);

            _model.Table.BlueUnits = new List<Unit>();
            _model.Table.RedUnits = new List<Unit>();
            Assert.AreEqual(0, _model.Table.BlueUnits.Count);
            Assert.AreEqual(0, _model.Table.RedUnits.Count);

            
            int blueGold = _model.Table.BlueGold;
            _model.Table.State = GameState.Blue;
            _model.BuyUnit(SelectedObject.LargeUnit);
            Assert.AreEqual(blueGold - LargeUnit.Price, _model.Table.BlueGold);
            Assert.AreEqual(1, _model.Table.BlueUnits.Count);

            

            int redGold = _model.Table.RedGold;
            _model.Table.State = GameState.Red;
            _model.BuyUnit(SelectedObject.SmallUnit);
            Assert.AreEqual(redGold - SmallUnit.Price, _model.Table.RedGold);
            Assert.AreEqual(1, _model.Table.RedUnits.Count);

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("BuyUnit"), TestCategory("Gameplay")]
        public async Task BuyUnitNotEnoughMoney()
        {
            await _model.LoadGameAsync(String.Empty);

            _model.Table.BlueUnits = new List<Unit>();
            _model.Table.RedUnits = new List<Unit>();
            Assert.AreEqual(0, _model.Table.BlueUnits.Count);
            Assert.AreEqual(0, _model.Table.RedUnits.Count);


            _model.Table.State = GameState.Blue;
            _model.Table.BlueGold = SmallUnit.Price - 1;
            _model.BuyUnit(SelectedObject.SmallUnit);
            Assert.AreEqual(SmallUnit.Price - 1, _model.Table.BlueGold);
            Assert.AreEqual(0, _model.Table.BlueUnits.Count);


            _model.Table.State = GameState.Red;
            _model.Table.RedGold = LargeUnit.Price - 1;
            _model.BuyUnit(SelectedObject.LargeUnit);
            Assert.AreEqual(LargeUnit.Price - 1, _model.Table.RedGold);
            Assert.AreEqual(0, _model.Table.RedUnits.Count);

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("MoveUnit"), TestCategory("Gameplay")]
        public async Task MoveBlueUnits()
        {
            await _model.LoadGameAsync(String.Empty);

            _model.Table.BlueUnits = new List<Unit>();

            _model.Table[0, 5] = Dune.Instance;
            _model.Table[1, 5] = Dune.Instance;
            _model.Table[6, 9] = Dune.Instance;

            //jobbra fog menni
            Unit a = new SmallUnit() { IsBlue = true, Y = 0, X = 1 };
            _model.Table.BlueUnits.Add(a);

            //lefel fog menni
            Unit b = new SmallUnit() { IsBlue = true, Y = 0, X = 4 };
            _model.Table.BlueUnits.Add(b);


            // balra fog menni
            Unit c = new LargeUnit() { IsBlue = true, Y = 7, X = 9 };
            _model.Table.BlueUnits.Add(c);


            // felfel fog menni
            Unit d = new LargeUnit() { IsBlue = true, Y = 2, X = 9 };
            _model.Table.BlueUnits.Add(d);


            _model.Table.BlueSmallUnits = 2;
            _model.Table.BlueLargeUnits = 2;
            _model.Table.BlueUnitsCurrentNum = 4;
            _model.Table.State = GameState.Red;
            _model.StepGameState(); // szimulci elindtsa, bels vltozk inicializlsa

            _model.Table.BlueUnitsCurrentPositions[0][1].Add(a);
            _model.Table.BlueUnitsCurrentPositions[0][4].Add(b);
            _model.Table.BlueUnitsCurrentPositions[7][9].Add(c);
            _model.Table.BlueUnitsCurrentPositions[2][9].Add(d);

            _model.Simulate(); // a szimulci egy lpsnek lejtszsa

            Assert.IsTrue(a.Y == 0 && a.X == 2);
            Assert.IsTrue(b.Y == 1 && b.X == 4);
            Assert.IsTrue(c.Y == 7 && c.X == 8);
            Assert.IsTrue(d.Y == 1 && d.X == 9);

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("MoveUnit"), TestCategory("Gameplay")]
        public async Task MoveRedUnits()
        {
            await _model.LoadGameAsync(String.Empty);

            _model.Table.RedUnits = new List<Unit>();

            //jobbra fog menni
            Unit a = new LargeUnit() { IsBlue = false, Y = 0, X = 1 };
            _model.Table.RedUnits.Add(a);

            //lefel fog menni
            Unit b = new LargeUnit() { IsBlue = false, Y = 0, X = 4 };
            _model.Table.RedUnits.Add(b);


            // balra fog menni
            Unit c = new SmallUnit() { IsBlue = false, Y = 4, X = 4 };
            _model.Table.RedUnits.Add(c);


            // felfel fog menni
            Unit d = new SmallUnit() { IsBlue = false, Y = 6, X = 5 };
            _model.Table.RedUnits.Add(d);


            _model.Table.RedSmallUnits = 2;
            _model.Table.RedLargeUnits = 2;
            _model.Table.RedUnitsCurrentNum = 4;
            _model.Table.State = GameState.Red;
            _model.StepGameState(); // szimulci elindtsa, bels vltozk inicializlsa

            _model.Table.RedUnitsCurrentPositions[0][1].Add(a);
            _model.Table.RedUnitsCurrentPositions[0][4].Add(b);
            _model.Table.RedUnitsCurrentPositions[4][4].Add(c);
            _model.Table.RedUnitsCurrentPositions[6][5].Add(d);

            _model.Simulate(); // a szimulci egy lpsnek lejtszsa

            Assert.IsTrue(a.Y == 0 && a.X == 2);
            Assert.IsTrue(b.Y == 1 && b.X == 4);
            Assert.IsTrue(c.Y == 4 && c.X == 3);
            Assert.IsTrue(d.Y == 5 && d.X == 5);

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("DamageCastle"), TestCategory("Gameplay")]
        public async Task DamageCastle()
        {
            await _model.LoadGameAsync(String.Empty);

            _model.Table.BlueUnits = new List<Unit>();

            Unit a = new SmallUnit() { IsBlue = true, Y = 1, X = 9 };
            _model.Table.BlueUnits.Add(a);

            
            Unit b = new SmallUnit() { IsBlue = true, Y = 4, X = 1 };
            _model.Table.RedUnits.Add(b);


            _model.Table.BlueSmallUnits = 1;
            _model.Table.RedSmallUnits = 1;
            _model.Table.BlueUnitsCurrentNum = 1;
            _model.Table.RedUnitsCurrentNum = 1;
            _model.Table.State = GameState.Red;
            _model.StepGameState(); // szimulci elindtsa, bels vltozk inicializlsa

            _model.Table.BlueUnitsCurrentPositions[1][9].Add(a);
            _model.Table.RedUnitsCurrentPositions[4][1].Add(b);

            Assert.AreEqual(10, _model.Table.BlueHealth);
            Assert.AreEqual(10, _model.Table.RedHealth);

            _model.Simulate(); // a szimulci egy lpsnek lejtszsa

            Assert.AreEqual(9, _model.Table.BlueHealth);
            Assert.AreEqual(9, _model.Table.RedHealth);

            // az egysgek meghalnak a tmads kvetkeztben
            Assert.AreEqual(0, _model.Table.BlueUnits.Count);
            Assert.AreEqual(0, _model.Table.RedUnits.Count);
            Assert.AreEqual(0, _model.Table.BlueUnitsCurrentNum);
            Assert.AreEqual(0, _model.Table.RedUnitsCurrentNum);

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("DamageCastle"), TestCategory("GameOver"), TestCategory("Gameplay")]
        public async Task BlueWin()
        {
            await _model.LoadGameAsync(String.Empty);

            _model.Table.BlueUnits = new List<Unit>();

            for (int i = 0; i < 10; ++i)
            {
                Unit a = new SmallUnit() { IsBlue = true, Y = 1, X = 9 };
                _model.Table.BlueUnits.Add(a);
            }


            _model.Table.BlueSmallUnits = 10;
            _model.Table.BlueUnitsCurrentNum = 10;
            _model.Table.State = GameState.Red;
            _model.StepGameState(); // szimulci elindtsa, bels vltozk inicializlsa

            for (int i = 0; i < 10; ++i)
            {
                _model.Table.BlueUnitsCurrentPositions[1][9].Add(_model.Table.BlueUnits[i]);
            }

            Assert.AreEqual(10, _model.Table.RedHealth);
            Assert.AreEqual(10, _model.Table.BlueUnits.Count);
            Assert.AreEqual(10, _model.Table.BlueUnitsCurrentNum);

            int signal = 0;
            _model.GameOver += new EventHandler<GameEventArgs>((obj, e) => signal = e.WhoWon);

            _model.Simulate(); // a szimulci egy lpsnek lejtszsa

            Assert.AreEqual(0, _model.Table.RedHealth);

            // az egysgek meghalnak a tmads kvetkeztben
            Assert.AreEqual(0, _model.Table.BlueUnits.Count);
            Assert.AreEqual(0, _model.Table.BlueUnitsCurrentNum);

            Assert.AreEqual(1, signal); // a kk nyert

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("DamageCastle"), TestCategory("GameOver"), TestCategory("Gameplay")]
        public async Task RedWin()
        {
            await _model.LoadGameAsync(String.Empty);

            _model.Table.RedUnits = new List<Unit>();

            for (int i = 0; i < 10; ++i)
            {
                Unit a = new SmallUnit() { IsBlue = false, Y = 4, X = 1 };
                _model.Table.RedUnits.Add(a);
            }


            _model.Table.RedSmallUnits = 10;
            _model.Table.RedUnitsCurrentNum = 10;
            _model.Table.State = GameState.Red;
            _model.StepGameState(); // szimulci elindtsa, bels vltozk inicializlsa

            for (int i = 0; i < 10; ++i)
            {
                _model.Table.RedUnitsCurrentPositions[4][1].Add(_model.Table.RedUnits[i]);
            }

            Assert.AreEqual(10, _model.Table.BlueHealth);
            Assert.AreEqual(10, _model.Table.RedUnits.Count);
            Assert.AreEqual(10, _model.Table.RedUnitsCurrentNum);

            int signal = 0;
            _model.GameOver += new EventHandler<GameEventArgs>((obj, e) => signal = e.WhoWon);

            _model.Simulate(); // a szimulci egy lpsnek lejtszsa

            Assert.AreEqual(0, _model.Table.BlueHealth);

            // az egysgek meghalnak a tmads kvetkeztben
            Assert.AreEqual(0, _model.Table.RedUnits.Count);
            Assert.AreEqual(0, _model.Table.RedUnitsCurrentNum);

            Assert.AreEqual(2, signal); // a piros nyert

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("DamageCastle"), TestCategory("GameOver"), TestCategory("Gameplay")]
        public async Task Draw()
        {
            await _model.LoadGameAsync(String.Empty);

            _model.Table.BlueUnits = new List<Unit>();

            for (int i = 0; i < 10; ++i)
            {
                Unit a = new SmallUnit() { IsBlue = true, Y = 1, X = 9 };
                _model.Table.BlueUnits.Add(a);
            }


            _model.Table.BlueSmallUnits = 10;
            _model.Table.BlueUnitsCurrentNum = 10;

            _model.Table.RedUnits = new List<Unit>();

            for (int i = 0; i < 10; ++i)
            {
                Unit a = new SmallUnit() { IsBlue = false, Y = 4, X = 1 };
                _model.Table.RedUnits.Add(a);
            }


            _model.Table.RedSmallUnits = 10;
            _model.Table.RedUnitsCurrentNum = 10;
            _model.Table.State = GameState.Red;
            _model.StepGameState(); // szimulci elindtsa, bels vltozk inicializlsa

            for (int i = 0; i < 10; ++i)
            {
                _model.Table.RedUnitsCurrentPositions[4][1].Add(_model.Table.RedUnits[i]);
            }

            for (int i = 0; i < 10; ++i)
            {
                _model.Table.BlueUnitsCurrentPositions[4][1].Add(_model.Table.BlueUnits[i]);
            }

            Assert.AreEqual(10, _model.Table.BlueHealth);
            Assert.AreEqual(10, _model.Table.RedUnits.Count);
            Assert.AreEqual(10, _model.Table.RedUnitsCurrentNum);
            Assert.AreEqual(10, _model.Table.RedHealth);
            Assert.AreEqual(10, _model.Table.BlueUnits.Count);
            Assert.AreEqual(10, _model.Table.BlueUnitsCurrentNum);

            int signal = 0;
            _model.GameOver += new EventHandler<GameEventArgs>((obj, e) => signal = e.WhoWon);

            _model.Simulate(); // a szimulci egy lpsnek lejtszsa

            Assert.AreEqual(0, _model.Table.BlueHealth);
            Assert.AreEqual(0, _model.Table.RedHealth);

            // az egysgek meghalnak a tmads kvetkeztben
            Assert.AreEqual(0, _model.Table.RedUnits.Count);
            Assert.AreEqual(0, _model.Table.RedUnitsCurrentNum);
            Assert.AreEqual(0, _model.Table.BlueUnits.Count);
            Assert.AreEqual(0, _model.Table.BlueUnitsCurrentNum);

            Assert.AreEqual(3, signal); // dntetlen

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("ShootUnit"), TestCategory("Gameplay")]
        public async Task ShootUnitsNotInRange()
        {
            await _model.LoadGameAsync(String.Empty);

            _model.Table.BlueUnits = new List<Unit>();

            _model.Table.State = GameState.Blue;
            _model.PlaceObject(1, 6, SelectedObject.SmallTower);
            _model.Table.State = GameState.Red;
            _model.PlaceObject(3, 9, SelectedObject.MediumTower);

            Unit a = new SmallUnit() { IsBlue = true, Y = 0, X = 1 };
            _model.Table.BlueUnits.Add(a);

            Unit b = new SmallUnit() { IsBlue = false, Y = 0, X = 2 };
            _model.Table.RedUnits.Add(b);


            _model.Table.BlueSmallUnits = 1;
            _model.Table.BlueUnitsCurrentNum = 1;
            _model.Table.RedSmallUnits = 1;
            _model.Table.RedUnitsCurrentNum = 1;

            _model.StepGameState(); // szimulci elindtsa, bels vltozk inicializlsa

            _model.Table.BlueUnitsCurrentPositions[0][1].Add(a);
            _model.Table.RedUnitsCurrentPositions[0][2].Add(b);

            int blueHealth = a.Health;
            int redHealth = b.Health;

            _model.Simulate(); // a szimulci egy lpsnek lejtszsa

            Assert.AreEqual(blueHealth, a.Health);
            Assert.AreEqual(redHealth, b.Health);

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("ShootUnit"), TestCategory("Gameplay")]
        public async Task ShootUnitsInRange()
        {
            await _model.LoadGameAsync(String.Empty);

            _model.Table.BlueUnits = new List<Unit>();

            _model.Table.State = GameState.Blue;
            _model.PlaceObject(1, 6, SelectedObject.SmallTower);
            _model.Table.State = GameState.Red;
            _model.PlaceObject(3, 9, SelectedObject.MediumTower);

            Unit a = new SmallUnit() { IsBlue = true, Y = 3, X = 8 };
            _model.Table.BlueUnits.Add(a);

            Unit b = new SmallUnit() { IsBlue = false, Y = 1, X = 8 };
            _model.Table.RedUnits.Add(b);


            _model.Table.BlueSmallUnits = 1;
            _model.Table.BlueUnitsCurrentNum = 1;
            _model.Table.RedSmallUnits = 1;
            _model.Table.RedUnitsCurrentNum = 1;

            _model.StepGameState(); // szimulci elindtsa, bels vltozk inicializlsa

            _model.Table.BlueUnitsCurrentPositions[3][8].Add(a);
            _model.Table.RedUnitsCurrentPositions[1][8].Add(b);

            int blueHealth = a.Health;
            int redHealth = b.Health;

            _model.Simulate(); // a szimulci egy lpsnek lejtszsa

            Assert.AreEqual(blueHealth - new MediumTower().Damage, a.Health);
            Assert.AreEqual(redHealth - new SmallTower().Damage, b.Health);

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }

        [TestMethod, TestCategory("ShootUnit"), TestCategory("Gameplay")]
        public async Task ShootUnitsDeath()
        {
            await _model.LoadGameAsync(String.Empty);

            _model.Table.BlueUnits = new List<Unit>();

            _model.Table.State = GameState.Blue;
            _model.PlaceObject(1, 6, SelectedObject.LargeTower);
            _model.PlaceObject(1, 5, SelectedObject.MediumTower);
            _model.Table.State = GameState.Red;
            _model.PlaceObject(3, 6, SelectedObject.LargeTower);
            _model.PlaceObject(3, 5, SelectedObject.MediumTower);

            Unit a = new SmallUnit() { IsBlue = true, Y = 2, X = 6 };
            _model.Table.BlueUnits.Add(a);

            Unit b = new SmallUnit() { IsBlue = false, Y = 2, X = 5 };
            _model.Table.RedUnits.Add(b);


            _model.Table.BlueSmallUnits = 1;
            _model.Table.BlueUnitsCurrentNum = 1;
            _model.Table.RedSmallUnits = 1;
            _model.Table.RedUnitsCurrentNum = 1;

            _model.StepGameState(); // szimulci elindtsa, bels vltozk inicializlsa

            _model.Table.BlueUnitsCurrentPositions[2][6].Add(a);
            _model.Table.RedUnitsCurrentPositions[2][5].Add(b);

            _model.Simulate(); // a szimulci egy lpsnek lejtszsa
            _model.Simulate();
            _model.Simulate();

            Assert.AreEqual(0, _model.Table.BlueUnitsCurrentNum);
            Assert.AreEqual(0, _model.Table.RedUnitsCurrentNum);

            _mock.Verify(dataAccess => dataAccess.LoadAsync(String.Empty), Times.Once());
        }
    }
}
